@@ -54,7 +54,7 @@ autoapi_type = "python"
 autoapi_dirs = ["../../slidge", "../../superduper"]
 autoapi_add_toctree_entry = False
 autoapi_keep_files = False
-autoapi_root = "dev/api"
+autoapi_root = "api/auto"
 autoapi_ignore = ["*xep_*", "slidge/core/*"]
 autoapi_options = [
     "members",
