@@ -158,7 +158,7 @@ class ListGroups(Command):
         assert session is not None
         await session.ready
         groups = sorted(
-            session.bookmarks, key=lambda g: (g.name or g.jid.localpart).casefold()
+            session.bookmarks, key=lambda g: (g.name or g.jid.node).casefold()
         )
         return TableResult(
             description="Your groups",
@@ -182,7 +182,7 @@ class Login(Command):
             raise XMPPError("bad-request", "You are already logging in.")
         session.is_logging_in = True
         try:
-            msg = await session.login()
+            msg = await self.xmpp.login_wrap(session)
         except Exception as e:
             session.send_gateway_status(f"Re-login failed: {e}", show="dnd")
             raise XMPPError(
