@@ -124,8 +124,10 @@ class LegacyRoster(
         if contact.stored.updated:
             return contact
 
-        with contact.updating_info():
+        with contact.updating_info(merge=False):
             await contact.update_info()
+            if contact.is_friend and not self.__filling:
+                await contact.add_to_roster()
 
         if contact.cached_presence is not None:
             contact._store_last_presence(contact.cached_presence)
@@ -161,10 +163,9 @@ class LegacyRoster(
             if self.get_lock(("username", username)):
                 self.log.debug("Already updating %s via by_jid()", username)
 
-                jid = JID()
-                jid.node = username
-                jid.domain = self.session.xmpp.boundjid.bare
-                return await self.by_jid(jid)
+                return await self.by_jid(
+                    JID(username + "@" + self.session.xmpp.boundjid.bare)
+                )
 
             with self.orm() as orm:
                 stored = (
@@ -234,6 +235,10 @@ class LegacyRoster(
                 self.log.debug("No need to update roster")
                 continue
             self.log.debug("Updating roster")
+            if not contact.is_friend:
+                continue
+            if not self.session.user.preferences.get("roster_push", True):
+                continue
             try:
                 await self.session.xmpp["xep_0356"].set_roster(
                     self.session.user_jid.bare,
