@@ -143,21 +143,35 @@ class MessageContentMixin(DispatcherMixin):
             return
 
         if isinstance(recipient, LegacyMUC):
-            await recipient.echo(msg, legacy_msg_id)
+            stanza_id = await recipient.echo(msg, legacy_msg_id)
         else:
+            stanza_id = None
             self.__ack(msg)
 
         if legacy_msg_id is None:
             return
 
         with self.xmpp.store.session() as orm:
-            self.xmpp.store.id_map.set_msg(
-                orm,
-                recipient.stored.id,
-                str(legacy_msg_id),
-                [msg.get_id()],
-                recipient.is_group,
-            )
+            if recipient.is_group:
+                self.xmpp.store.id_map.set_origin(
+                    orm, recipient.stored.id, str(legacy_msg_id), msg.get_id()
+                )
+                assert stanza_id is not None
+                self.xmpp.store.id_map.set_msg(
+                    orm,
+                    recipient.stored.id,
+                    str(legacy_msg_id),
+                    [stanza_id],
+                    True,
+                )
+            else:
+                self.xmpp.store.id_map.set_msg(
+                    orm,
+                    recipient.stored.id,
+                    str(legacy_msg_id),
+                    [msg.get_id()],
+                    False,
+                )
             if session.MESSAGE_IDS_ARE_THREAD_IDS and (t := msg["thread"]):
                 self.xmpp.store.id_map.set_thread(
                     orm, recipient.stored.id, t, str(legacy_msg_id), recipient.is_group
@@ -170,18 +184,9 @@ class MessageContentMixin(DispatcherMixin):
             # ignore message retraction fallback (fallback=last msg correction)
             return
         session, recipient, thread = await self._get_session_recipient_thread(msg)
-        xmpp_id = msg["replace"]["id"]
-        if isinstance(recipient, LegacyMUC):
-            with self.xmpp.store.session() as orm:
-                legacy_id_str = self.xmpp.store.id_map.get_legacy(
-                    orm, recipient.stored.id, xmpp_id, True
-                )
-            if legacy_id_str is None:
-                legacy_id = self._xmpp_msg_id_to_legacy(session, xmpp_id, recipient)
-            else:
-                legacy_id = self.xmpp.LEGACY_MSG_ID_TYPE(legacy_id_str)
-        else:
-            legacy_id = self._xmpp_msg_id_to_legacy(session, xmpp_id, recipient)
+        legacy_id = self._xmpp_msg_id_to_legacy(
+            session, msg["replace"]["id"], recipient, True
+        )
 
         if isinstance(recipient, LegacyMUC):
             mentions = await recipient.parse_mentions(msg["body"])
@@ -246,7 +251,7 @@ class MessageContentMixin(DispatcherMixin):
                 orm,
                 recipient.stored.id,
                 new_legacy_msg_id,
-                msg.get_id(),
+                [msg.get_id()],
                 recipient.is_group,
             )
             orm.commit()
@@ -260,13 +265,12 @@ class MessageContentMixin(DispatcherMixin):
                 "This legacy service does not support message retraction.",
             )
         xmpp_id: str = msg["retract"]["id"]
-        legacy_id = self._xmpp_msg_id_to_legacy(session, xmpp_id, recipient)
-        if legacy_id:
-            await session.on_retract(recipient, legacy_id, thread=thread)
-            if isinstance(recipient, LegacyMUC):
-                await recipient.echo(msg, None)
-        else:
-            log.debug("Ignored retraction from user")
+        legacy_id = self._xmpp_msg_id_to_legacy(
+            session, xmpp_id, recipient, origin=True
+        )
+        await session.on_retract(recipient, legacy_id, thread=thread)
+        if isinstance(recipient, LegacyMUC):
+            await recipient.echo(msg, None)
         self.__ack(msg)
 
     @exceptions_to_xmpp_errors
@@ -358,7 +362,7 @@ class MessageContentMixin(DispatcherMixin):
             session.log.debug(
                 "Could not determine reply-to legacy msg ID, sending quote instead."
             )
-            return msg["body"], None, None, None
+            return redact_url(msg["body"]), None, None, None
 
         reply_to_jid = JID(msg["reply"]["to"])
         reply_to = None
@@ -375,17 +379,19 @@ class MessageContentMixin(DispatcherMixin):
             except XMPPError:
                 pass
             else:
-                if nick != muc.user_nick:
-                    reply_to = await muc.get_participant(
-                        reply_to_jid.resource, store=False
-                    )
+                if nick == muc.user_nick:
+                    reply_to = await muc.get_user_participant()
+                elif not nick:
+                    reply_to = muc.get_system_participant()
+                else:
+                    reply_to = await muc.get_participant(nick, store=False)
 
         if msg.get_plugin("fallback", check=True) and (
             isinstance(recipient, LegacyMUC) or recipient.REPLIES
         ):
             text = msg["fallback"].get_stripped_body(self.xmpp["xep_0461"].namespace)
             try:
-                reply_fallback = msg["reply"].get_fallback_body()
+                reply_fallback = redact_url(msg["reply"].get_fallback_body())
             except AttributeError:
                 reply_fallback = None
         else:
@@ -425,4 +431,11 @@ class MessageContentMixin(DispatcherMixin):
         return await session.on_sticker(recipient, sticker, **kwargs)
 
 
+def redact_url(text: str) -> str:
+    needle = config.NO_UPLOAD_URL_PREFIX or config.UPLOAD_URL_PREFIX
+    if not needle:
+        return text
+    return text.replace(needle, "")
+
+
 log = logging.getLogger(__name__)
