@@ -6,7 +6,8 @@ from slixmpp import JID, Iq, Message, Presence
 from slixmpp.exceptions import XMPPError
 from slixmpp.xmlstream import StanzaBase
 
-from ...util.types import Recipient, RecipientType
+from ...contact.roster import ContactIsUser
+from ...util.types import LegacyMessageType, Recipient, RecipientType
 from ..session import BaseSession
 
 if TYPE_CHECKING:
@@ -75,11 +76,15 @@ class DispatcherMixin:
         return muc
 
     def _xmpp_msg_id_to_legacy(
-        self, session: "BaseSession", xmpp_id: str, recipient: Recipient
-    ):
+        self,
+        session: "BaseSession[LegacyMessageType, Any]",
+        xmpp_id: str,
+        recipient: Recipient,
+        origin: bool = False,
+    ) -> LegacyMessageType:
         with self.xmpp.store.session() as orm:
             sent = self.xmpp.store.id_map.get_legacy(
-                orm, recipient.stored.id, xmpp_id, False
+                orm, recipient.stored.id, xmpp_id, recipient.is_group, origin
             )
             if sent is not None:
                 return self.xmpp.LEGACY_MSG_ID_TYPE(sent)
@@ -176,6 +181,10 @@ def exceptions_to_xmpp_errors(cb: HandlerType) -> HandlerType:
                 f"{cb.__name__} is not implemented by the legacy module",
                 clear=False,
             )
+        except ContactIsUser:
+            raise XMPPError(
+                "bad-request", "Actions with your bridged self are not allowed."
+            )
         except Exception as e:
             log.error("Failed to handle incoming stanza: %s", args, exc_info=e)
             raise XMPPError("internal-server-error", str(e))
