@@ -17,6 +17,8 @@ from slixmpp import JID, ComponentXMPP, Iq, Message, Presence
 from slixmpp.exceptions import IqError, IqTimeout, XMPPError
 from slixmpp.plugins.xep_0004 import Form
 from slixmpp.plugins.xep_0060.stanza import OwnerAffiliation
+from slixmpp.plugins.xep_0356.permissions import IqPermission
+from slixmpp.plugins.xep_0356.privilege import PrivilegedIqError
 from slixmpp.types import MessageTypes
 from slixmpp.xmlstream.xmlstream import NotConnectedError
 
@@ -29,13 +31,13 @@ from slidge.command.register import RegistrationType
 from slidge.core import config
 from slidge.core.dispatcher.session_dispatcher import SessionDispatcher
 from slidge.core.mixins import MessageMixin
+from slidge.core.mixins.attachment import AttachmentMixin
 from slidge.core.mixins.avatar import convert_avatar
 from slidge.core.pubsub import PubSubComponent
 from slidge.core.session import BaseSession
 from slidge.db import GatewayUser, SlidgeStore
 from slidge.db.avatar import CachedAvatar, avatar_cache
 from slidge.db.meta import JSONSerializable
-from slidge.slixfix import PrivilegedIqError
 from slidge.slixfix.delivery_receipt import DeliveryReceipt
 from slidge.slixfix.roster import RosterBackend
 from slidge.util import ABCSubclassableOnceAtMost
@@ -283,6 +285,14 @@ class BaseGateway(
     Common example: ``int``.
     """
 
+    DB_POOL_SIZE: int = 5
+    """
+    Size of the queue pool for sqlalchemy engine. Typically, when using python async
+    libraries, this does not need to be changed.
+    Change that if your gateway use separate threads to call into slidge. The value of
+    this parameter should be equal or greater than the potential number of threads.
+    """
+
     http: aiohttp.ClientSession
     avatar: CachedAvatar | None = None
 
@@ -327,6 +337,12 @@ class BaseGateway(
         self.has_crashed: bool = False
         self.use_origin_id = False
 
+        if config.USER_JID_VALIDATOR is None:
+            config.USER_JID_VALIDATOR = f".*@{self.infer_real_domain()}"
+            log.info(
+                "No USER_JID_VALIDATOR was set, using '%s'.",
+                config.USER_JID_VALIDATOR,
+            )
         self.jid_validator: re.Pattern = re.compile(config.USER_JID_VALIDATOR)
         self.qr_pending_registrations = dict[str, asyncio.Future[Optional[dict]]]()
 
@@ -465,7 +481,8 @@ class BaseGateway(
         )
         self.del_event_handler("presence_probe", self._handle_probe)
         self.add_event_handler("session_start", self.__on_session_start)
-        self.add_event_handler("disconnected", self.connect)
+        self.add_event_handler("privileges_advertised", self.__on_privileges_advertised)
+        self.__upload_service_found = asyncio.Event()
 
     def __register_slixmpp_api(self) -> None:
         def with_session(func, commit: bool = True):
@@ -496,6 +513,8 @@ class BaseGateway(
     async def __on_session_start(self, event) -> None:
         log.debug("Gateway session start: %s", event)
 
+        await self.__setup_attachments()
+
         # prevents XMPP clients from considering the gateway as an HTTP upload
         disco = self.plugin["xep_0030"]
         await disco.del_feature(feature="urn:xmpp:http:upload:0", jid=self.boundjid)
@@ -543,6 +562,32 @@ class BaseGateway(
 
         log.info("Slidge has successfully started")
 
+    async def __on_privileges_advertised(self, _) -> None:
+        await self.__upload_service_found.wait()
+        if self.xmpp["xep_0356"].granted_privileges[config.SERVER].iq.get(
+            self.plugin["xep_0363"].stanza.Request.namespace
+        ) in (IqPermission.SET, IqPermission.BOTH, IqPermission.GET):
+            AttachmentMixin.PRIVILEGED_UPLOAD = True
+
+    async def __setup_attachments(self) -> None:
+        if config.NO_UPLOAD_PATH:
+            if config.NO_UPLOAD_URL_PREFIX is None:
+                raise RuntimeError(
+                    "If you set NO_UPLOAD_PATH you must set NO_UPLOAD_URL_PREFIX too."
+                )
+        elif not config.UPLOAD_SERVICE:
+            info_iq = await self.xmpp.plugin["xep_0363"].find_upload_service(
+                self.infer_real_domain()
+            )
+            if info_iq is None:
+                raise RuntimeError("Could not find upload service")
+            log.info("Auto-discovered upload service: %s", info_iq["from"])
+            config.UPLOAD_SERVICE = info_iq["from"]
+        self.__upload_service_found.set()
+
+    def infer_real_domain(self) -> JID:
+        return JID(re.sub(r"^.*?\.", "", self.xmpp.boundjid.bare))
+
     async def __add_component_to_mds_whitelist(self, user_jid: JID) -> None:
         # Uses privileged entity to add ourselves to the whitelist of the PEP
         # MDS node so we receive MDS events
@@ -597,21 +642,22 @@ class BaseGateway(
             )
 
     @timeit
-    async def login_wrap(self, session: "BaseSession") -> None:
+    async def login_wrap(self, session: "BaseSession") -> str:
         session.send_gateway_status("Logging in…", show="dnd")
         session.is_logging_in = True
         try:
             status = await session.login()
         except Exception as e:
             log.warning("Login problem for %s", session.user_jid, exc_info=e)
-            log.exception(e)
             session.send_gateway_status(f"Could not login: {e}", show="busy")
-            session.send_gateway_message(
+            msg = (
                 "You are not connected to this gateway! "
                 f"Maybe this message will tell you why: {e}"
             )
+            session.send_gateway_message(msg)
             session.logged = False
-            return
+            session.send_gateway_status("Login failed", show="dnd")
+            return msg
 
         log.info("Login success for %s", session.user_jid)
         session.logged = True
@@ -627,9 +673,9 @@ class BaseGateway(
                 r.set_result(True)
         self.send_presence(pto=session.user.jid.bare, ptype="probe")
         if status is None:
-            session.send_gateway_status("Logged in", show="chat")
-        else:
-            session.send_gateway_status(status, show="chat")
+            status = "Logged in"
+        session.send_gateway_status(status, show="chat")
+
         if session.user.preferences.get("sync_avatar", False):
             session.create_task(self.fetch_user_avatar(session))
         else:
@@ -637,6 +683,7 @@ class BaseGateway(
                 session.user.avatar_hash = None
                 orm.add(session.user)
                 orm.commit()
+        return status
 
     async def fetch_user_avatar(self, session: BaseSession) -> None:
         try:
@@ -678,7 +725,8 @@ class BaseGateway(
         if not self.jid_validator.match(jid.bare):
             raise XMPPError(
                 condition="not-allowed",
-                text="Your account is not allowed to use this gateway.",
+                text="Your account is not allowed to use this gateway. "
+                "The admin controls that with the USER_JID_VALIDATOR option.",
             )
 
     def send_raw(self, data: Union[str, bytes]):
@@ -742,17 +790,6 @@ class BaseGateway(
             self.loop.stop()
             exit(1)
 
-    def re_login(self, session: "BaseSession") -> None:
-        async def w() -> None:
-            session.cancel_all_tasks()
-            try:
-                await session.logout()
-            except NotImplementedError:
-                pass
-            await self.login_wrap(session)
-
-        session.create_task(w())
-
     async def make_registration_form(
         self, _jid: JID, _node: str, _ifrom: JID, iq: Iq
     ) -> Form:
@@ -914,12 +951,10 @@ class BaseGateway(
         else:
             fut.set_exception(exception)
 
-    async def unregister_user(self, user: GatewayUser) -> None:
-        self.send_presence(
-            pshow="dnd",
-            pstatus="You unregistered from this gateway.",
-            pto=user.jid,
-        )
+    async def unregister_user(
+        self, user: GatewayUser, msg: str = "You unregistered from this gateway."
+    ) -> None:
+        self.send_presence(pshow="dnd", pstatus=msg, pto=user.jid)
         await self.xmpp.plugin["xep_0077"].api["user_remove"](None, None, user.jid)
         await self.xmpp._session_cls.kill_by_jid(user.jid)
 
