@@ -2,11 +2,9 @@ import base64
 import functools
 import logging
 import os
-import re
 import shutil
 import stat
 import tempfile
-import warnings
 from datetime import datetime
 from itertools import chain
 from mimetypes import guess_extension, guess_type
@@ -16,12 +14,11 @@ from urllib.parse import quote as urlquote
 from uuid import uuid4
 from xml.etree import ElementTree as ET
 
+import aiohttp
 import thumbhash
 from PIL import Image, ImageOps
-from slixmpp import JID, Message
-from slixmpp.exceptions import IqError, IqTimeout
+from slixmpp import JID, Iq, Message
 from slixmpp.plugins.xep_0264.stanza import Thumbnail
-from slixmpp.plugins.xep_0363 import FileUploadError
 from slixmpp.plugins.xep_0447.stanza import StatelessFileSharing
 
 from ...db.avatar import avatar_cache
@@ -38,48 +35,71 @@ from .message_text import TextMessageMixin
 
 
 class AttachmentMixin(TextMessageMixin):
-    is_group: bool
+    PRIVILEGED_UPLOAD = False
+
+    @property
+    def __is_component(self) -> bool:
+        return self.session is NotImplemented
 
     async def __upload(
         self,
         file_path: Path,
         file_name: Optional[str] = None,
         content_type: Optional[str] = None,
-    ) -> str | None:
-        if file_name and file_path.name != file_name:
-            d = Path(tempfile.mkdtemp())
-            temp = d / file_name
-            temp.symlink_to(file_path)
-            file_path = temp
-        else:
-            d = None
-        if config.UPLOAD_SERVICE:
-            domain = None
-        else:
-            domain = re.sub(r"^.*?\.", "", self.xmpp.boundjid.bare)
-        try:
-            new_url = await self.xmpp.plugin["xep_0363"].upload_file(
-                filename=file_path,
-                content_type=content_type,
-                ifrom=config.UPLOAD_REQUESTER or self.xmpp.boundjid,
-                domain=JID(domain),
-            )
-        except (FileUploadError, IqError, IqTimeout) as e:
-            warnings.warn(f"Something is wrong with the upload service: {e!r}")
-            return None
-        finally:
-            if d is not None:
-                file_path.unlink()
-                d.rmdir()
+    ) -> str:
+        assert config.UPLOAD_SERVICE
 
-        return new_url
+        file_size = file_path.stat().st_size
+
+        content_type = content_type or self.xmpp.plugin["xep_0363"].default_content_type
+        iq_slot = await self.__request_upload_slot(
+            config.UPLOAD_SERVICE,
+            file_name or file_path.name,
+            file_size,
+            content_type,
+        )
+        slot = iq_slot["http_upload_slot"]
+        headers = {
+            "Content-Length": str(file_size),
+            "Content-Type": content_type,
+            **{header["name"]: header["value"] for header in slot["put"]["headers"]},
+        }
+
+        async with aiohttp.ClientSession() as http:
+            with file_path.open("rb") as fp:
+                async with http.put(
+                    slot["put"]["url"], data=fp, headers=headers
+                ) as response:
+                    response.raise_for_status()
+        return slot["get"]["url"]
+
+    async def __request_upload_slot(
+        self,
+        upload_service: JID | str,
+        filename: str,
+        size: int,
+        content_type: str,
+    ) -> Iq:
+        iq_request = self.xmpp.make_iq_get(
+            ito=upload_service, ifrom=config.UPLOAD_REQUESTER or self.xmpp.boundjid
+        )
+        request = iq_request["http_upload_request"]
+        request["filename"] = filename
+        request["size"] = str(size)
+        request["content-type"] = content_type
+        if self.__is_component or not self.PRIVILEGED_UPLOAD:
+            return await iq_request.send()
+
+        assert self.session is not NotImplemented
+        iq_request.set_from(self.session.user_jid)
+        return await self.xmpp["xep_0356"].send_privileged_iq(iq_request)
 
     @staticmethod
     async def __no_upload(
         file_path: Path,
         file_name: Optional[str] = None,
         legacy_file_id: Optional[Union[str, int]] = None,
-    ):
+    ) -> tuple[Path, str]:
         file_id = str(uuid4()) if legacy_file_id is None else str(legacy_file_id)
         assert config.NO_UPLOAD_PATH is not None
         assert config.NO_UPLOAD_URL_PREFIX is not None
@@ -141,7 +161,7 @@ class AttachmentMixin(TextMessageMixin):
             return r.status < 400
 
     async def __get_stored(self, attachment: LegacyAttachment) -> Attachment:
-        if attachment.legacy_file_id is not None and self.session is not NotImplemented:
+        if attachment.legacy_file_id is not None and not self.__is_component:
             with self.xmpp.store.session() as orm:
                 stored = (
                     orm.query(Attachment)
@@ -156,9 +176,7 @@ class AttachmentMixin(TextMessageMixin):
                         stored.url = None  # type:ignore
                     return stored
         return Attachment(
-            user_account_id=None
-            if self.session is NotImplemented
-            else self.session.user_pk,
+            user_account_id=None if self.__is_component else self.session.user_pk,
             legacy_file_id=None
             if attachment.legacy_file_id is None
             else str(attachment.legacy_file_id),
@@ -167,7 +185,7 @@ class AttachmentMixin(TextMessageMixin):
 
     async def __get_url(
         self, attachment: LegacyAttachment, stored: Attachment
-    ) -> tuple[bool, Optional[Path], str]:
+    ) -> tuple[bool, Path | None, str]:
         file_name = attachment.name
         content_type = attachment.content_type
         file_path = attachment.path
@@ -233,6 +251,9 @@ class AttachmentMixin(TextMessageMixin):
         if stored.legacy_file_id and new_url is not None:
             stored.url = new_url
 
+        if local_path is not None and local_path.stat().st_size == 0:
+            raise RuntimeError("File size is 0")
+
         return is_temp, local_path, new_url
 
     async def __set_sims(
@@ -372,14 +393,13 @@ class AttachmentMixin(TextMessageMixin):
             msg["fallback"].enable("body")
             msg["fallback"]["for"] = self.xmpp.plugin["xep_0447"].stanza.NAMESPACE
         if caption:
-            m1 = self._send(msg, carbon=carbon, **kwargs)
+            if correction:
+                msg["replace"]["id"] = self._replace_id(legacy_msg_id)
+            else:
+                self._set_msg_id(msg, legacy_msg_id)
+            m1 = self._send(msg, carbon=carbon, correction=correction, **kwargs)
             m2 = self.send_text(
-                caption,
-                legacy_msg_id=legacy_msg_id,
-                when=when,
-                carbon=carbon,
-                correction=correction,
-                **kwargs,
+                caption, legacy_msg_id=None, when=when, carbon=carbon, **kwargs
             )
             return [m1, m2] if m2 else [m1]
         else:
@@ -399,16 +419,14 @@ class AttachmentMixin(TextMessageMixin):
         correction: bool = False,
         mto: Optional[JID] = None,
     ) -> Message:
-        if correction and (original_xmpp_id := self._legacy_to_xmpp(legacy_msg_id)):
-            with self.xmpp.store.session() as orm:
-                xmpp_ids = self.xmpp.store.id_map.get_xmpp(
-                    orm, self._recipient_pk(), str(legacy_msg_id), self.is_group
-                )
-
-            for xmpp_id in xmpp_ids:
-                if xmpp_id == original_xmpp_id:
-                    continue
-                self.retract(xmpp_id, thread)
+        if correction:
+            xmpp_ids = self._legacy_to_xmpp(legacy_msg_id)
+            if xmpp_ids:
+                original_xmpp_id = xmpp_ids[0]
+                for xmpp_id in xmpp_ids:
+                    if xmpp_id == original_xmpp_id:
+                        continue
+                    self.retract(xmpp_id, thread)
 
         if reply_to is not None and reply_to.body:
             # We cannot have a "quote fallback" for attachments since most (all?)
@@ -461,7 +479,7 @@ class AttachmentMixin(TextMessageMixin):
             thread=thread,
             **kwargs,
         )
-        if self.session is NotImplemented:
+        if self.__is_component:
             return await coro
         elif not isinstance(attachment, LegacyAttachment):
             return await coro
@@ -510,17 +528,18 @@ class AttachmentMixin(TextMessageMixin):
             local_path = None
             new_url = stored.url
         else:
-            is_temp, local_path, new_url = await self.__get_url(attachment, stored)
-            if new_url is None or (
-                local_path is not None and local_path.stat().st_size == 0
-            ):
-                log.warning("Something went wrong with this attachment: %s", attachment)
+            try:
+                is_temp, local_path, new_url = await self.__get_url(attachment, stored)
+            except Exception as e:
+                log.error("Error with attachment: %s: %s", attachment, e)
+                log.debug("", exc_info=e)
                 msg["body"] = (
-                    "I tried to send a file, but something went wrong. "
-                    "Tell your slidge admin to check the logs."
+                    f"/me tried to send a file ({attachment.format_for_user()}), "
+                    f"but something went wrong: {e}. "
                 )
                 self._set_msg_id(msg, legacy_msg_id)
                 return None, [self._send(msg, **kwargs)]
+        assert new_url is not None
 
         stored.url = new_url
         if config.USE_ATTACHMENT_ORIGINAL_URLS and attachment.url:
@@ -531,7 +550,7 @@ class AttachmentMixin(TextMessageMixin):
             )
             self.__set_sfs(msg, new_url, local_path, attachment, stored, thumbnail)
 
-        if self.session is not NotImplemented:
+        if not self.__is_component:
             with self.xmpp.store.session(expire_on_commit=False) as orm:
                 orm.add(stored)
                 orm.commit()
@@ -543,7 +562,7 @@ class AttachmentMixin(TextMessageMixin):
         msgs = self.__send_url(
             msg, legacy_msg_id, new_url, attachment.caption, carbon, when, **kwargs
         )
-        if self.session is not NotImplemented:
+        if not self.__is_component:
             if store_multi:
                 self.__store_multi(legacy_msg_id, msgs)
         return new_url, msgs
@@ -590,6 +609,9 @@ class AttachmentMixin(TextMessageMixin):
         if not attachments and not body:
             # ignoring empty message
             return
+        body_msg_id = (
+            legacy_msg_id if body_needs_msg_id(attachments, body, body_first) else None
+        )
         send_body = functools.partial(
             self.__send_body,
             body=body,
@@ -597,17 +619,15 @@ class AttachmentMixin(TextMessageMixin):
             when=when,
             thread=thread,
             correction=correction,
-            legacy_msg_id=legacy_msg_id,
+            legacy_msg_id=body_msg_id,
             correction_event_id=correction_event_id,
             **kwargs,
         )
         all_msgs = []
         if body_first:
             all_msgs.append(send_body())
-        last_attachment_i = len(attachments) - 1
         for i, attachment in enumerate(attachments):
-            last = i == last_attachment_i
-            if last and not body:
+            if i == 0 and body_msg_id is None:
                 legacy = legacy_msg_id
             else:
                 legacy = None
@@ -642,11 +662,20 @@ class AttachmentMixin(TextMessageMixin):
                 ids.append(msg.get_id())
         with self.xmpp.store.session() as orm:
             self.xmpp.store.id_map.set_msg(
-                orm, self._recipient_pk(), str(legacy_msg_id), ids, self.is_group
+                orm, self._recipient_pk(), str(legacy_msg_id), ids, self.is_participant
             )
             orm.commit()
 
 
+def body_needs_msg_id(
+    attachments: Collection[LegacyAttachment], body: str | None, body_first: bool
+) -> bool:
+    if attachments:
+        return bool(body and body_first)
+    else:
+        return True
+
+
 def get_thumbhash(path: Path) -> tuple[str, int, int]:
     with path.open("rb") as fp:
         img = Image.open(fp)
