@@ -31,7 +31,7 @@ class Base:
 
 
 class BaseSender(Base):
-    is_participant = False
+    is_participant: bool = NotImplemented
 
     def _send(
         self, stanza: MessageOrPresenceTypeVar, **send_kwargs
