@@ -3,8 +3,9 @@ import typing
 from contextlib import contextmanager
 
 import sqlalchemy as sa
+from sqlalchemy.exc import IntegrityError
 
-from ...db.models import Base, Contact, Room
+from ...db.models import Base, Contact, Participant, Room
 
 if typing.TYPE_CHECKING:
     from slidge import BaseGateway
@@ -75,12 +76,37 @@ class UpdateInfoMixin(DBMixin):
         pass
 
     @contextmanager
-    def updating_info(self):
+    def updating_info(self, merge=True):
         self._updating_info = True
         yield
         self._updating_info = False
         self.stored.updated = True
-        self.commit()
+        try:
+            self.commit(merge=merge)
+        except IntegrityError as e:
+            from slidge.group import LegacyMUC
+
+            if not isinstance(self, LegacyMUC):
+                raise e
+            if not self._ALL_INFO_FILLED_ON_STARTUP:
+                raise e
+            with self.xmpp.store.session(expire_on_commit=False) as orm:
+                if self.stored.id is None:
+                    self.stored.id = self.xmpp.store.rooms.get(
+                        orm, self.user_pk, legacy_id=str(self.legacy_id)
+                    ).id
+                merged = orm.merge(self.stored)
+                resources = set()
+                participants: list[Participant] = []
+                for participant in merged.participants:
+                    if participant.resource in resources:
+                        self.session.log.warning("ditching: %s", participant)
+                        continue
+                    resources.add(participant.resource)
+                merged.participants = participants
+                self.stored = merged
+                orm.add(self.stored)
+                orm.commit()
 
     def commit(self, merge: bool = False) -> None:
         if self._updating_info:
