@@ -16,7 +16,7 @@ import aiohttp
 import sqlalchemy as sa
 from slixmpp import JID, Message
 from slixmpp.exceptions import XMPPError
-from slixmpp.types import PresenceShows
+from slixmpp.types import PresenceShows, ResourceDict
 
 from ..command import SearchResult
 from ..contact import LegacyContact, LegacyRoster
@@ -33,7 +33,6 @@ from ..util.types import (
     Mention,
     PseudoPresenceShow,
     RecipientType,
-    ResourceDict,
     Sticker,
 )
 from ..util.util import deprecated, noop_coro
@@ -566,6 +565,11 @@ class BaseSession(
                 return
             self.__reset_ready()
             self.shutdown(logout=False)
+            with self.xmpp.store.session() as orm:
+                self.xmpp.store.mam.reset_source(orm)
+                self.xmpp.store.rooms.reset_updated(orm)
+                self.xmpp.store.contacts.reset_updated(orm)
+                orm.commit()
         else:
             if v:
                 self.ready.set_result(True)
@@ -597,6 +601,8 @@ class BaseSession(
             await self.logout()
         except NotImplementedError:
             pass
+        except KeyboardInterrupt:
+            pass
 
     @staticmethod
     def legacy_to_xmpp_msg_id(legacy_msg_id: LegacyMessageType) -> str:
@@ -807,12 +813,6 @@ class BaseSession(
         """
         await self.xmpp.send_qr(text, mto=self.user_jid)
 
-    def re_login(self) -> None:
-        # Logout then re-login
-        #
-        # No reason to override this
-        self.xmpp.re_login(self)
-
     async def get_contact_or_group_or_participant(self, jid: JID, create: bool = True):
         if (contact := self.contacts.by_jid_only_if_exists(jid)) is not None:
             return contact
@@ -837,12 +837,7 @@ class BaseSession(
     @staticmethod
     async def __get_muc_or_participant(muc: LegacyMUC, jid: JID):
         if nick := jid.resource:
-            try:
-                return await muc.get_participant(
-                    nick, raise_if_not_found=True, fill_first=True
-                )
-            except XMPPError:
-                return None
+            return await muc.get_participant(nick, create=False, fill_first=True)
         return muc
 
     async def wait_for_ready(self, timeout: Optional[Union[int, float]] = 10):
