@@ -248,6 +248,7 @@ class ArchivedMessage(Base):
 
     id: Mapped[int] = mapped_column(primary_key=True)
     room_id: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)
+    room: Mapped[Room] = relationship(lazy=True, back_populates="archive")
 
     stanza_id: Mapped[str] = mapped_column(nullable=False)
     timestamp: Mapped[datetime] = mapped_column(nullable=False)
@@ -257,6 +258,8 @@ class ArchivedMessage(Base):
 
     stanza: Mapped[str] = mapped_column(nullable=False)
 
+    displayed_by_user: Mapped[bool] = mapped_column(default=False)
+
 
 class _LegacyToXmppIdsBase:
     """
@@ -282,6 +285,20 @@ class GroupMessages(_LegacyToXmppIdsBase, Base):
     foreign_key: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)
 
 
+class GroupMessagesOrigin(_LegacyToXmppIdsBase, Base):
+    """
+    This maps "origin ids" <message id=XXX> to legacy message IDs
+    We need that for message corrections and retractions, which do not reference
+    messages by their "Unique and Stable Stanza IDs (XEP-0359)"
+    """
+
+    __tablename__ = "group_msg_origin"
+    __table_args__ = (
+        Index("ix_group_msg_origin_legacy_id", "legacy_id", "foreign_key"),
+    )
+    foreign_key: Mapped[int] = mapped_column(ForeignKey("room.id"), nullable=False)
+
+
 class DirectThreads(_LegacyToXmppIdsBase, Base):
     __tablename__ = "direct_thread"
     __table_args__ = (Index("ix_direct_direct_thread_id", "legacy_id", "foreign_key"),)
@@ -317,6 +334,7 @@ class Participant(Base):
     __table_args__ = (
         UniqueConstraint("room_id", "resource"),
         UniqueConstraint("room_id", "contact_id"),
+        UniqueConstraint("room_id", "occupant_id"),
     )
 
     id: Mapped[int] = mapped_column(primary_key=True)
@@ -333,10 +351,14 @@ class Participant(Base):
         lazy=False, back_populates="participants"
     )
 
+    occupant_id: Mapped[str] = mapped_column(nullable=False)
+
     is_user: Mapped[bool] = mapped_column(default=False)
 
-    affiliation: Mapped[MucAffiliation] = mapped_column(default="member")
-    role: Mapped[MucRole] = mapped_column(default="participant")
+    affiliation: Mapped[MucAffiliation] = mapped_column(
+        default="member", nullable=False
+    )
+    role: Mapped[MucRole] = mapped_column(default="participant", nullable=False)
 
     presence_sent: Mapped[bool] = mapped_column(default=False)
 
@@ -348,6 +370,11 @@ class Participant(Base):
 
     extra_attributes: Mapped[Optional[JSONSerializable]] = mapped_column(default=None)
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.role = "participant"
+        self.affiliation = "member"
+
 
 class Bob(Base):
     __tablename__ = "bob"
