@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import hashlib
 import logging
+import shutil
 import uuid
 from datetime import datetime, timedelta, timezone
 from mimetypes import guess_extension
@@ -29,6 +30,7 @@ from .models import (
     DirectThreads,
     GatewayUser,
     GroupMessages,
+    GroupMessagesOrigin,
     GroupThreads,
     Participant,
     Room,
@@ -39,12 +41,15 @@ class UpdatedMixin:
     model: Type[Base] = NotImplemented
 
     def __init__(self, session: Session) -> None:
-        session.execute(update(self.model).values(updated=False))
+        self.reset_updated(session)
 
     def get_by_pk(self, session: Session, pk: int) -> Type[Base]:
         stmt = select(self.model).where(self.model.id == pk)  # type:ignore
         return session.scalar(stmt)
 
+    def reset_updated(self, session: Session) -> None:
+        session.execute(update(self.model).values(updated=False))
+
 
 class SlidgeStore:
     def __init__(self, engine: Engine) -> None:
@@ -89,6 +94,7 @@ LegacyToXmppType = (
     | Type[DirectThreads]
     | Type[GroupMessages]
     | Type[GroupThreads]
+    | Type[GroupMessagesOrigin]
 )
 
 
@@ -146,6 +152,27 @@ class IdMapStore:
             GroupMessages if group else DirectMessages,
         )
 
+    def set_origin(
+        self, session: Session, foreign_key: int, legacy_id: str, xmpp_id: str
+    ) -> None:
+        self._set(
+            session,
+            foreign_key,
+            legacy_id,
+            [xmpp_id],
+            GroupMessagesOrigin,
+        )
+
+    def get_origin(
+        self, session: Session, foreign_key: int, legacy_id: str
+    ) -> list[str]:
+        return self._get(
+            session,
+            foreign_key,
+            legacy_id,
+            GroupMessagesOrigin,
+        )
+
     @staticmethod
     def _get(
         session: Session, foreign_key: int, legacy_id: str, type_: LegacyToXmppType
@@ -177,8 +204,20 @@ class IdMapStore:
         )
 
     def get_legacy(
-        self, session: Session, foreign_key: int, xmpp_id: str, group: bool
+        self,
+        session: Session,
+        foreign_key: int,
+        xmpp_id: str,
+        group: bool,
+        origin: bool = False,
     ) -> Optional[str]:
+        if origin and group:
+            return self._get_legacy(
+                session,
+                foreign_key,
+                xmpp_id,
+                GroupMessagesOrigin,
+            )
         return self._get_legacy(
             session,
             foreign_key,
@@ -250,6 +289,10 @@ class ContactStore(UpdatedMixin):
 class MAMStore:
     def __init__(self, session: Session, session_maker) -> None:
         self.session = session_maker
+        self.reset_source(session)
+
+    @staticmethod
+    def reset_source(session: Session) -> None:
         session.execute(
             update(ArchivedMessage).values(source=ArchivedMessageSource.BACKFILL)
         )
@@ -455,12 +498,63 @@ class MAMStore:
             .first()
         )
 
+    @staticmethod
+    def pop_unread_up_to(session: Session, room_pk: int, stanza_id: str) -> list[str]:
+        q = (
+            select(ArchivedMessage.id, ArchivedMessage.stanza_id)
+            .where(ArchivedMessage.room_id == room_pk)
+            .where(~ArchivedMessage.displayed_by_user)
+            .where(ArchivedMessage.legacy_id.is_not(None))
+            .order_by(ArchivedMessage.timestamp.asc())
+        )
+
+        ref = session.scalar(
+            select(ArchivedMessage)
+            .where(ArchivedMessage.room_id == room_pk)
+            .where(ArchivedMessage.stanza_id == stanza_id)
+        )
+
+        if ref is None:
+            log.debug(
+                "(pop unread in muc): message not found, returning all MAM messages."
+            )
+            rows = session.execute(q)
+        else:
+            rows = session.execute(q.where(ArchivedMessage.timestamp <= ref.timestamp))
+
+        pks: list[int] = []
+        stanza_ids: list[str] = []
+
+        for id_, stanza_id in rows:
+            pks.append(id_)
+            stanza_ids.append(stanza_id)
+
+        session.execute(
+            update(ArchivedMessage)
+            .where(ArchivedMessage.id.in_(pks))
+            .values(displayed_by_user=True)
+        )
+        return stanza_ids
+
+    @staticmethod
+    def is_displayed_by_user(
+        session: Session, room_jid: str, legacy_msg_id: str
+    ) -> bool:
+        return any(
+            session.execute(
+                select(ArchivedMessage.displayed_by_user)
+                .join(Room)
+                .where(Room.jid == room_jid)
+                .where(ArchivedMessage.legacy_id == legacy_msg_id)
+            ).scalars()
+        )
+
 
 class RoomStore(UpdatedMixin):
     model = Room
 
-    def __init__(self, session: Session) -> None:
-        super().__init__(session)
+    def reset_updated(self, session: Session) -> None:
+        super().reset_updated(session)
         session.execute(
             update(Room).values(
                 subject_setter=None,
@@ -474,6 +568,22 @@ class RoomStore(UpdatedMixin):
     def get_all(session: Session, user_pk: int) -> Iterator[Room]:
         yield from session.scalars(select(Room).where(Room.user_account_id == user_pk))
 
+    @staticmethod
+    def get(session: Session, user_pk: int, legacy_id: str) -> Room:
+        return session.execute(
+            select(Room)
+            .where(Room.user_account_id == user_pk)
+            .where(Room.legacy_id == legacy_id)
+        ).scalar_one()
+
+    @staticmethod
+    def nick_available(session: Session, room_pk: int, nickname: str) -> bool:
+        return (
+            session.execute(
+                select(Participant.id).filter_by(room_id=room_pk, nickname=nickname)
+            )
+        ).one_or_none() is None
+
 
 class ParticipantStore:
     def __init__(self, session: Session) -> None:
@@ -506,7 +616,11 @@ class BobStore:
     }
 
     def __init__(self) -> None:
-        self.root_dir = config.HOME_DIR / "slidge_stickers"
+        if (config.HOME_DIR / "slidge_stickers").exists():
+            shutil.move(
+                config.HOME_DIR / "slidge_stickers", config.HOME_DIR / "bob_store"
+            )
+        self.root_dir = config.HOME_DIR / "bob_store"
         self.root_dir.mkdir(exist_ok=True)
 
     @staticmethod
@@ -614,7 +728,7 @@ def _check_avatar_orphans(session, flush_context):
         )
     )
     deleted_count = result.rowcount
-    log.debug(f"Auto-deleted %s orphaned avatars", deleted_count)
+    log.debug("Auto-deleted %s orphaned avatars", deleted_count)
 
 
 log = logging.getLogger(__name__)
