@@ -4,10 +4,10 @@ import uuid
 import warnings
 from copy import copy
 from datetime import datetime
-from functools import cached_property
 from typing import TYPE_CHECKING, Any, Optional, Union
 from xml.etree import ElementTree as ET
 
+import sqlalchemy as sa
 from slixmpp import JID, InvalidJID, Message, Presence
 from slixmpp.plugins.xep_0045.stanza import MUCAdminItem
 from slixmpp.types import MessageTypes, OptJid
@@ -117,10 +117,17 @@ class LegacyParticipant(
         #        But it does not happen for LegacyContact. WTF?
         raise RuntimeError
 
-    def commit(self, *args, **kwargs) -> None:
+    def __should_commit(self) -> bool:
         if self.is_system:
-            return
-        if self.muc.get_lock("fill participants") or self.muc.get_lock("fill history"):
+            return False
+        if self.muc.get_lock("fill participants"):
+            return False
+        if self.muc.get_lock("fill history"):
+            return False
+        return True
+
+    def commit(self, *args, **kwargs) -> None:
+        if not self.__should_commit():
             return
         super().commit(*args, **kwargs)
 
@@ -144,7 +151,15 @@ class LegacyParticipant(
         if self._presence_sent == val:
             return
         self.stored.presence_sent = val
-        self.commit(merge=True)
+        if not self.__should_commit():
+            return
+        with self.xmpp.store.session() as orm:
+            orm.execute(
+                sa.update(Participant)
+                .where(Participant.id == self.stored.id)
+                .values(presence_sent=val)
+            )
+            orm.commit()
 
     @property
     def nickname_no_illegal(self) -> str:
@@ -247,6 +262,16 @@ class LegacyParticipant(
         if new_nickname == old:
             return
 
+        if self.muc.stored.id is not None:
+            with self.xmpp.store.session() as orm:
+                if not self.xmpp.store.rooms.nick_available(
+                    orm, self.muc.stored.id, new_nickname
+                ):
+                    if self.contact is None:
+                        new_nickname = f"{new_nickname} ({self.occupant_id})"
+                    else:
+                        new_nickname = f"{new_nickname} ({self.contact.legacy_id})"
+
         cache = getattr(self, "_last_presence", None)
         if cache:
             last_seen = cache.last_seen
@@ -331,16 +356,9 @@ class LegacyParticipant(
                 return
             self.send_initial_presence(full_jid)
 
-    @cached_property
-    def __occupant_id(self):
-        if self.contact:
-            return self.contact.jid
-        elif self.is_user:
-            return "slidge-user"
-        elif self.is_system:
-            return "room"
-        else:
-            return str(uuid.uuid4())
+    @property
+    def occupant_id(self) -> str:
+        return self.stored.occupant_id
 
     def _send(
         self,
@@ -352,7 +370,7 @@ class LegacyParticipant(
         **send_kwargs,
     ) -> MessageOrPresenceTypeVar:
         if stanza.get_from().resource:
-            stanza["occupant-id"]["id"] = self.__occupant_id
+            stanza["occupant-id"]["id"] = self.occupant_id
         else:
             stanza["occupant-id"]["id"] = "room"
         self.__add_nick_element(stanza)
@@ -396,7 +414,7 @@ class LegacyParticipant(
             else:
                 warnings.warn(
                     (
-                        f"Public group but no contact JID associated to {self.jid} in"
+                        f"Private group but no contact JID associated to {self.jid} in"
                         f" {self}"
                     ),
                 )
@@ -486,20 +504,14 @@ class LegacyParticipant(
     def moderate(
         self, legacy_msg_id: LegacyMessageType, reason: Optional[str] = None
     ) -> None:
-        xmpp_id = self._legacy_to_xmpp(legacy_msg_id)
-        with self.xmpp.store.session() as orm:
-            msg_ids = self.xmpp.store.id_map.get_xmpp(
-                orm, self.muc.stored.id, str(legacy_msg_id), True
-            )
-
-        for i in set(msg_ids + [xmpp_id]):
+        for i in self._legacy_to_xmpp(legacy_msg_id):
             m = self.muc.get_system_participant()._make_message()
             m["retract"]["id"] = i
             if self.is_system:
                 m["retract"].enable("moderated")
             else:
                 m["retract"]["moderated"]["by"] = self.jid
-                m["retract"]["moderated"]["occupant-id"]["id"] = self.__occupant_id
+                m["retract"]["moderated"]["occupant-id"]["id"] = self.occupant_id
             if reason:
                 m["retract"]["reason"] = reason
             self._send(m)
