@@ -2,6 +2,7 @@ import json
 import logging
 import re
 import string
+import uuid
 import warnings
 from asyncio import Lock
 from contextlib import asynccontextmanager
@@ -9,15 +10,16 @@ from copy import copy
 from datetime import datetime, timedelta, timezone
 from typing import (
     TYPE_CHECKING,
+    Any,
     AsyncIterator,
     Generic,
+    Iterator,
     Literal,
     Optional,
     Type,
     Union,
     overload,
 )
-from uuid import uuid4
 
 import sqlalchemy as sa
 from slixmpp import JID, Iq, Message, Presence
@@ -29,6 +31,7 @@ from slixmpp.plugins.xep_0469.stanza import NS as PINNING_NS
 from slixmpp.plugins.xep_0492.stanza import NS as NOTIFY_NS
 from slixmpp.plugins.xep_0492.stanza import WhenLiteral
 from slixmpp.xmlstream import ET
+from sqlalchemy.exc import IntegrityError
 from sqlalchemy.orm import Session as OrmSession
 
 from ..contact.contact import LegacyContact
@@ -37,6 +40,7 @@ from ..core.mixins.avatar import AvatarMixin
 from ..core.mixins.disco import ChatterDiscoMixin
 from ..core.mixins.recipient import ReactionRecipientMixin, ThreadRecipientMixin
 from ..db.models import Participant, Room
+from ..util.archive_msg import HistoryMessage
 from ..util.jid_escaping import unescape_node
 from ..util.types import (
     HoleBound,
@@ -134,7 +138,33 @@ class LegacyMUC(
         self._set_logger()
         super().__init__()
 
-        self.archive = MessageArchive(stored, self.xmpp.store.mam)
+        self.archive = MessageArchive(stored, self.xmpp.store)
+
+        if self._ALL_INFO_FILLED_ON_STARTUP:
+            self.stored.participants_filled = True
+
+    def pop_unread_xmpp_ids_up_to(self, horizon_xmpp_id: str) -> list[str]:
+        """
+        Return XMPP msg ids sent in this group up to a given XMPP msg id.
+
+        Plugins have no reason to use this, but it is used by slidge core
+        for legacy networks that need to mark *all* messages as read (most XMPP
+        clients only send a read marker for the latest message).
+
+        This has side effects: all messages up to the horizon XMPP id will be marked
+        as read in the DB. If the horizon XMPP id is not found, all messages of this
+        MUC will be marked as read.
+
+        :param horizon_xmpp_id: The latest message
+        :return: A list of XMPP ids if horizon_xmpp_id was not found
+        """
+        with self.xmpp.store.session() as orm:
+            assert self.stored.id is not None
+            ids = self.xmpp.store.mam.pop_unread_up_to(
+                orm, self.stored.id, horizon_xmpp_id
+            )
+            orm.commit()
+        return ids
 
     def participant_from_store(
         self, stored: Participant, contact: LegacyContact | None = None
@@ -158,8 +188,8 @@ class LegacyMUC(
     def legacy_id(self):
         return self.xmpp.LEGACY_ROOM_ID_TYPE(self.stored.legacy_id)
 
-    def orm(self) -> OrmSession:
-        return self.xmpp.store.session()
+    def orm(self, **kwargs) -> OrmSession:
+        return self.xmpp.store.session(**kwargs)
 
     @property
     def type(self) -> MucType:
@@ -254,6 +284,9 @@ class LegacyMUC(
         return self.session.get_lock((self.legacy_id, id_))
 
     async def __fill_participants(self) -> None:
+        if self._ALL_INFO_FILLED_ON_STARTUP or self.participants_filled:
+            return
+
         async with self.lock("fill participants"):
             with self.xmpp.store.session(expire_on_commit=False) as orm:
                 orm.add(self.stored)
@@ -267,8 +300,8 @@ class LegacyMUC(
             # return a participant with a conflicting nick/resource.
             async for participant in self.fill_participants():
                 if participant.stored.resource in resources:
-                    self.log.warning(
-                        "Participant '%s' was yielded more than once by fill_participants()",
+                    self.log.debug(
+                        "Participant '%s' was yielded more than once by fill_participants(), ignoring",
                         participant.stored.resource,
                     )
                     continue
@@ -291,7 +324,7 @@ class LegacyMUC(
     ) -> AsyncIterator[LegacyParticipantType]:
         await self.__fill_participants()
         with self.xmpp.store.session(expire_on_commit=False, autoflush=False) as orm:
-            orm.add(self.stored)
+            self.stored = orm.merge(self.stored)
             for db_participant in self.stored.participants:
                 if (
                     affiliation is not None
@@ -460,7 +493,10 @@ class LegacyMUC(
     def __get_subject_setter_participant(self) -> LegacyParticipant:
         if self.subject_setter is None:
             return self.get_system_participant()
-        return self._participant_cls(self, Participant(nickname=self.subject_setter))
+        return self._participant_cls(
+            self,
+            Participant(nickname=self.subject_setter, occupant_id="subject-setter"),
+        )
 
     def features(self):
         features = [
@@ -566,7 +602,7 @@ class LegacyMUC(
 
     async def echo(
         self, msg: Message, legacy_msg_id: Optional[LegacyMessageType] = None
-    ) -> None:
+    ) -> str:
         origin_id = msg.get_origin_id()
 
         msg.set_from(self.user_muc_jid)
@@ -578,11 +614,13 @@ class LegacyMUC(
         if legacy_msg_id:
             msg["stanza_id"]["id"] = self.session.legacy_to_xmpp_msg_id(legacy_msg_id)
         else:
-            msg["stanza_id"]["id"] = str(uuid4())
+            msg["stanza_id"]["id"] = str(uuid.uuid4())
         msg["stanza_id"]["by"] = self.jid
-        msg["occupant-id"]["id"] = "slidge-user"
 
-        self.archive.add(msg, await self.get_user_participant())
+        user_part = await self.get_user_participant()
+        msg["occupant-id"]["id"] = user_part.stored.occupant_id
+
+        self.archive.add(msg, user_part)
 
         for user_full_jid in self.user_full_jids():
             self.log.debug("Echoing to %s", user_full_jid)
@@ -591,6 +629,8 @@ class LegacyMUC(
 
             msg.send()
 
+        return msg["stanza_id"]["id"]
+
     def _post_avatar_update(self, cached_avatar) -> None:
         self.__send_configuration_change((104,))
         self._send_room_presence()
@@ -702,16 +742,93 @@ class LegacyMUC(
     def __store_participant(self, p: "LegacyParticipantType") -> None:
         if self.get_lock("fill participants"):
             return
-        p.commit(merge=True)
+        try:
+            p.commit(merge=True)
+        except IntegrityError as e:
+            if self._ALL_INFO_FILLED_ON_STARTUP:
+                log.debug("ℂould not store participant: %r", e)
+                with self.orm(expire_on_commit=False) as orm:
+                    self.stored = self.xmpp.store.rooms.get(
+                        orm, self.user_pk, legacy_id=str(self.legacy_id)
+                    )
+                    p.stored.room = self.stored
+                    orm.add(p.stored)
+                    orm.commit()
+            else:
+                log.debug("ℂould not store participant: %r", e)
 
+    @overload
+    async def get_participant(self, nickname: str) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant(self, *, occupant_id: str) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant(
+        self, *, occupant_id: str, create: Literal[False]
+    ) -> "LegacyParticipantType | None": ...
+
+    @overload
+    async def get_participant(
+        self, *, occupant_id: str, create: Literal[True]
+    ) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant(
+        self, nickname: str, *, occupant_id: str
+    ) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant(
+        self, nickname: str, *, create: Literal[False]
+    ) -> "LegacyParticipantType | None": ...
+
+    @overload
+    async def get_participant(
+        self, nickname: str, *, create: Literal[True]
+    ) -> "LegacyParticipantType": ...
+
+    @overload
     async def get_participant(
         self,
         nickname: str,
-        raise_if_not_found: bool = False,
+        *,
+        create: Literal[True],
+        is_user: bool,
+        fill_first: bool,
+        store: bool,
+    ) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant(
+        self,
+        nickname: str,
+        *,
+        create: Literal[False],
+        is_user: bool,
+        fill_first: bool,
+        store: bool,
+    ) -> "LegacyParticipantType | None": ...
+
+    @overload
+    async def get_participant(
+        self,
+        nickname: str,
+        *,
+        create: bool,
+        fill_first: bool,
+    ) -> "LegacyParticipantType | None": ...
+
+    async def get_participant(
+        self,
+        nickname: str | None = None,
+        *,
+        create: bool = True,
+        is_user: bool = False,
         fill_first: bool = False,
         store: bool = True,
-        is_user: bool = False,
-    ) -> "LegacyParticipantType":
+        occupant_id: str | None = None,
+    ) -> "LegacyParticipantType | None":
         """
         Get a participant by their nickname.
 
@@ -719,33 +836,75 @@ class LegacyMUC(
         :meth:`.LegacyMUC.get_participant_by_contact` instead.
 
         :param nickname: Nickname of the participant (used as resource part in the MUC)
-        :param raise_if_not_found: Raise XMPPError("item-not-found") if they are not
-            in the participant list (internal use by slidge, plugins should not
-            need that)
-        :param fill_first: Ensure :meth:`.LegacyMUC.fill_participants()` has been called first
-             (internal use by slidge, plugins should not need that)
+        :param create: By default, a participant is created if necessary. Set this to
+            False to return None if participant was not created before.
+        :param is_user: Whether this participant is the slidge user.
+        :param fill_first: Ensure :meth:`.LegacyMUC.fill_participants()` has been called
+            first (internal use by slidge, plugins should not need that)
         :param store: persistently store the user in the list of MUC participants
-        :return:
+        :param occupant_id: optionally, specify the unique ID for this participant, cf
+            xep:`0421`
+        :return: A participant of this room.
         """
+        if not any((nickname, occupant_id)):
+            raise TypeError("You must specify either a nickname or an occupant ID")
         if fill_first:
             await self.__fill_participants()
-        with self.xmpp.store.session() as orm:
-            stored = (
-                orm.query(Participant)
-                .filter(
-                    Participant.room == self.stored,
-                    (Participant.nickname == nickname)
-                    | (Participant.resource == nickname),
-                )
-                .one_or_none()
-            )
-            if stored is not None:
-                return self.participant_from_store(stored)
+        if not self._ALL_INFO_FILLED_ON_STARTUP or self.stored.id is not None:
+            with self.xmpp.store.session(expire_on_commit=False) as orm:
+                if occupant_id is not None:
+                    stored = (
+                        orm.query(Participant)
+                        .filter(
+                            Participant.room == self.stored,
+                            Participant.occupant_id == occupant_id,
+                        )
+                        .one_or_none()
+                    )
+                elif nickname is not None:
+                    stored = (
+                        orm.query(Participant)
+                        .filter(
+                            Participant.room == self.stored,
+                            (Participant.nickname == nickname)
+                            | (Participant.resource == nickname),
+                        )
+                        .one_or_none()
+                    )
+                else:
+                    raise RuntimeError("NEVER")
+                if stored is not None:
+                    if occupant_id and occupant_id != stored.occupant_id:
+                        warnings.warn(
+                            f"Occupant ID mismatch in get_participant(): {occupant_id} vs {stored.occupant_id}",
+                        )
+                    part = self.participant_from_store(stored)
+                    if occupant_id and nickname and nickname != stored.nickname:
+                        stored.nickname = nickname
+                        orm.add(stored)
+                        orm.commit()
+                    return part
+
+        if not create:
+            return None
+
+        if occupant_id is None:
+            occupant_id = "slidge-user" if is_user else str(uuid.uuid4())
+
+        if nickname is None:
+            nickname = occupant_id
+
+        if not self.xmpp.store.rooms.nick_available(orm, self.stored.id, nickname):
+            nickname = f"{nickname} ({occupant_id})"
 
-        if raise_if_not_found:
-            raise XMPPError("item-not-found")
         p = self._participant_cls(
-            self, Participant(room=self.stored, nickname=nickname, is_user=is_user)
+            self,
+            Participant(
+                room=self.stored,
+                nickname=nickname or occupant_id,
+                is_user=is_user,
+                occupant_id=occupant_id,
+            ),
         )
         if store:
             self.__store_participant(p)
@@ -768,25 +927,40 @@ class LegacyMUC(
         service
         :return:
         """
-        return self._participant_cls(self, Participant(), is_system=True)
+        return self._participant_cls(
+            self, Participant(occupant_id="room"), is_system=True
+        )
 
     @overload
     async def get_participant_by_contact(
-        self, c: "LegacyContact"
+        self, c: "LegacyContact[Any]"
     ) -> "LegacyParticipantType": ...
 
     @overload
     async def get_participant_by_contact(
-        self, c: "LegacyContact", create: Literal[False]
+        self, c: "LegacyContact[Any]", *, occupant_id: str | None = None
+    ) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant_by_contact(
+        self,
+        c: "LegacyContact[Any]",
+        *,
+        create: Literal[False],
+        occupant_id: str | None,
     ) -> "LegacyParticipantType | None": ...
 
     @overload
     async def get_participant_by_contact(
-        self, c: "LegacyContact", create: Literal[True]
+        self,
+        c: "LegacyContact[Any]",
+        *,
+        create: Literal[True],
+        occupant_id: str | None,
     ) -> "LegacyParticipantType": ...
 
     async def get_participant_by_contact(
-        self, c: "LegacyContact", create: bool = True
+        self, c: "LegacyContact", *, create: bool = True, occupant_id: str | None = None
     ) -> "LegacyParticipantType | None":
         """
         Get a non-anonymous participant.
@@ -796,22 +970,29 @@ class LegacyMUC(
 
         :param c: The :class:`.LegacyContact` instance corresponding to this contact
         :param create: Creates the participant if it does not exist.
+        :param occupant_id: Optionally, specify a unique occupant ID (:xep:`0421`) for
+            this participant.
         :return:
         """
         await self.session.contacts.ready
 
-        with self.xmpp.store.session() as orm:
-            self.stored = orm.merge(self.stored)
-            stored = (
-                orm.query(Participant)
-                .filter_by(contact=c.stored, room=self.stored)
-                .one_or_none()
-            )
-            if stored is None:
-                if not create:
-                    return None
-            else:
-                return self.participant_from_store(stored=stored, contact=c)
+        if not self._ALL_INFO_FILLED_ON_STARTUP or self.stored.id is not None:
+            with self.xmpp.store.session() as orm:
+                self.stored = orm.merge(self.stored)
+                stored = (
+                    orm.query(Participant)
+                    .filter_by(contact=c.stored, room=self.stored)
+                    .one_or_none()
+                )
+                if stored is None:
+                    if not create:
+                        return None
+                else:
+                    if occupant_id and stored.occupant_id != occupant_id:
+                        warnings.warn(
+                            f"Occupant ID mismatch: {occupant_id} vs {stored.occupant_id}",
+                        )
+                    return self.participant_from_store(stored=stored, contact=c)
 
         nickname = c.name or unescape_node(c.jid.node)
 
@@ -819,17 +1000,21 @@ class LegacyMUC(
             nick_available = True
         else:
             with self.xmpp.store.session() as orm:
-                nick_available = (
-                    orm.query(Participant.id).filter_by(
-                        room=self.stored, nickname=nickname
-                    )
-                ).one_or_none() is None
+                nick_available = self.xmpp.store.rooms.nick_available(
+                    orm, self.stored.id, nickname
+                )
 
         if not nick_available:
             self.log.debug("Nickname conflict")
             nickname = f"{nickname} ({c.jid.node})"
         p = self._participant_cls(
-            self, Participant(nickname=nickname, room=self.stored), contact=c
+            self,
+            Participant(
+                nickname=nickname,
+                room=self.stored,
+                occupant_id=occupant_id or str(c.jid),
+            ),
+            contact=c,
         )
 
         self.__store_participant(p)
@@ -846,14 +1031,43 @@ class LegacyMUC(
             p.send_last_presence(force=True, no_cache_online=True)
         return p
 
+    @overload
     async def get_participant_by_legacy_id(
         self, legacy_id: LegacyUserIdType
+    ) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant_by_legacy_id(
+        self,
+        legacy_id: LegacyUserIdType,
+        *,
+        occupant_id: str | None,
+        create: Literal[True],
+    ) -> "LegacyParticipantType": ...
+
+    @overload
+    async def get_participant_by_legacy_id(
+        self,
+        legacy_id: LegacyUserIdType,
+        *,
+        occupant_id: str | None,
+        create: Literal[False],
+    ) -> "LegacyParticipantType | None": ...
+
+    async def get_participant_by_legacy_id(
+        self,
+        legacy_id: LegacyUserIdType,
+        *,
+        occupant_id: str | None = None,
+        create: bool = True,
     ) -> "LegacyParticipantType":
         try:
             c = await self.session.contacts.by_legacy_id(legacy_id)
         except ContactIsUser:
-            return await self.get_user_participant()
-        return await self.get_participant_by_contact(c)
+            return await self.get_user_participant(occupant_id=occupant_id)
+        return await self.get_participant_by_contact(  # type:ignore[call-overload]
+            c, create=create, occupant_id=occupant_id
+        )
 
     def remove_participant(
         self,
@@ -1062,7 +1276,7 @@ class LegacyMUC(
             item = ans["pubsub"]["items"]["item"]
             item["id"] = self.jid
             return item
-        except IqTimeout as exc:
+        except IqTimeout:
             warnings.warn(f"Cannot fetch bookmark for {self.user_jid}: timeout")
             return None
         except IqError as exc:
@@ -1344,6 +1558,23 @@ class LegacyMUC(
     def participants_filled(self) -> bool:
         return self.stored.participants_filled
 
+    def get_archived_messages(
+        self, msg_id: LegacyMessageType | str
+    ) -> Iterator[HistoryMessage]:
+        """
+        Query the slidge archive for messages sent in this group
+
+        :param msg_id: Message ID of the message in question. Can be either a legacy ID
+            or an XMPP ID.
+        :return: Iterator over messages. A single legacy ID can map to several messages,
+            because of multi-attachment messages.
+        """
+        with self.xmpp.store.session() as orm:
+            for stored in self.xmpp.store.mam.get_messages(
+                orm, self.stored.id, ids=[str(msg_id)]
+            ):
+                yield HistoryMessage(stored.stanza)
+
 
 def set_origin_id(msg: Message, origin_id: str) -> None:
     sub = ET.Element("{urn:xmpp:sid:0}origin-id")
