@@ -2,7 +2,7 @@
 Typing stuff
 """
 
-from dataclasses import dataclass
+from dataclasses import dataclass, fields
 from datetime import datetime
 from enum import IntEnum
 from pathlib import Path
@@ -22,7 +22,7 @@ from typing import (
 )
 
 from slixmpp import Message, Presence
-from slixmpp.types import PresenceShows, PresenceTypes, ResourceDict
+from slixmpp.types import PresenceShows, PresenceTypes, ResourceDict  # noqa: F401
 
 if TYPE_CHECKING:
     from ..contact import LegacyContact
@@ -131,6 +131,36 @@ class LegacyAttachment:
         if isinstance(self.path, str):
             self.path = Path(self.path)
 
+    def format_for_user(self) -> str:
+        if self.name:
+            name = self.name
+        elif self.path:
+            name = self.path.name  # type:ignore[union-attr]
+        elif self.url:
+            name = self.url
+        else:
+            name = ""
+
+        if self.caption:
+            if name:
+                name = f"{name}: {self.caption}"
+            else:
+                name = self.caption
+
+        return name
+
+    def __str__(self):
+        attrs = ", ".join(
+            f"{f.name}={getattr(self, f.name)!r}"
+            for f in fields(self)
+            if getattr(self, f.name) is not None and f.name != "data"
+        )
+        if self.data is not None:
+            data_str = f"data=<{len(self.data)} bytes>"
+            to_join = (attrs, data_str) if attrs else (data_str,)
+            attrs = ", ".join(to_join)
+        return f"Attachment({attrs})"
+
 
 class MucType(IntEnum):
     """
@@ -178,8 +208,7 @@ class Mention(NamedTuple):
 class Hat(NamedTuple):
     uri: str
     title: str
-    hue: float = 0.0  # Default value is not great here, but an OK workaround for the
-    # slixmpp 1.11→1.12 API update
+    hue: float | None = None
 
 
 class UserPreferences(TypedDict):
