@@ -1,13 +1,12 @@
 import logging
 import mimetypes
 import re
-import subprocess
 import warnings
 from abc import ABCMeta
 from functools import wraps
 from pathlib import Path
 from time import time
-from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Optional, Type, TypeVar
+from typing import Any, Callable, NamedTuple, Optional, Type, TypeVar
 
 try:
     import emoji
@@ -16,10 +15,9 @@ except ImportError:
 else:
     EMOJI_LIB_AVAILABLE = True
 
-from .types import Mention, ResourceDict
+from slixmpp.types import ResourceDict
 
-if TYPE_CHECKING:
-    from ..contact.contact import LegacyContact
+from .types import Mention
 
 try:
     import magic
@@ -125,8 +123,8 @@ def is_valid_phone_number(phone: Optional[str]):
     return match[0] == phone
 
 
-def strip_illegal_chars(s: str):
-    return ILLEGAL_XML_CHARS_RE.sub("", s)
+def strip_illegal_chars(s: str, repl: str = "") -> str:
+    return ILLEGAL_XML_CHARS_RE.sub(repl, s)
 
 
 # from https://stackoverflow.com/a/64570125/5902284 and Link Mauve
@@ -288,7 +286,7 @@ def dict_to_named_tuple(data: dict, cls: Type[T]) -> T:
 def replace_mentions(
     text: str,
     mentions: Optional[list[Mention]],
-    mapping: Callable[["LegacyContact"], str],
+    mapping: Callable[[Mention], str],
 ):
     if not mentions:
         return text
@@ -296,7 +294,12 @@ def replace_mentions(
     cursor = 0
     pieces = []
     for mention in mentions:
-        pieces.extend([text[cursor : mention.start], mapping(mention.contact)])
+        try:
+            new_text = mapping(mention)
+        except Exception as exc:
+            log.debug("Attempting slidge <= 0.3.3 compatibility: %s", exc)
+            new_text = mapping(mention.contact)  # type:ignore
+        pieces.extend([text[cursor : mention.start], new_text])
         cursor = mention.end
     pieces.append(text[cursor:])
     return "".join(pieces)
