@@ -1,6 +1,7 @@
 import asyncio
 import hashlib
 import io
+import unittest
 from base64 import b64encode
 from contextlib import asynccontextmanager
 from http import HTTPStatus
@@ -88,3 +89,28 @@ class AvatarFixtureMixin:
     avatar_original_sha1: str
     avatar_url: str
     avatar_base64: str
+
+
+class MockUUID4:
+    def __init__(self, i: int) -> None:
+        self.i = i
+        self.hex = str(i)
+
+    def __repr__(self):
+        return self.i.__repr__()
+
+
+class UUIDFixtureMixin(unittest.TestCase):
+    def _next_uuid(self):
+        self._uuid_counter += 1
+        return self._uuid_counter
+
+    def setUp(self):
+        super().setUp()
+        self._uuid_counter = 0
+        self._uuid_patch = unittest.mock.patch("uuid.uuid4", side_effect=self._next_uuid)
+        self._uuid_patch.start()
+
+    def tearDown(self) -> None:
+        super().tearDown()
+        self._uuid_patch.stop()
