@@ -9,6 +9,8 @@ from pathlib import Path
 from unittest.mock import ANY, MagicMock, patch
 
 import pytest
+from slixmpp.plugins.xep_0356.permissions import MessagePermission
+
 from conftest import AvatarFixtureMixin
 from test_shakespeare import Base as Shakespeare
 
@@ -36,12 +38,14 @@ def attachment(request, avatar_path):
             yield MockHeadResponse
 
         @asynccontextmanager
-        async def get(*a, **k):
+        async def get(url, *a, **k):
+            if "fail" in url:
+                raise RuntimeError("IT FAILED")
             yield MockGetResponse
 
     with (
         patch(
-            "slixmpp.plugins.xep_0363.http_upload.XEP_0363.upload_file",
+            "slidge.core.mixins.attachment.AttachmentMixin._AttachmentMixin__upload",
             return_value="http://url",
         ) as http_upload,
         patch("aiohttp.ClientSession", return_value=MockAioHTTP) as client_session,
@@ -172,21 +176,23 @@ class TestBodyOnly(Base):
 
 
 class TestAttachmentUpload(Base):
-    def __test_basic(self, attachment: LegacyAttachment, upload_kwargs: dict, disposition: str = ''):
+    def __test_basic(
+        self, attachment: LegacyAttachment, upload_args: tuple, disposition: str = ""
+    ):
         """
         Basic test that file is uploaded.
         """
         self.run_coro(self.juliet.send_files([attachment]))
-        self.http_upload.assert_called_with(**upload_kwargs)
+        self.http_upload.assert_called_with(*upload_args)
         self._assert_file(disposition=disposition)
 
-    def _test_reuse(self, attachment: LegacyAttachment, upload_kwargs: dict):
+    def _test_reuse(self, attachment: LegacyAttachment, upload_args: tuple):
         """
         Basic test the no new file is uploaded when the same attachment is used
         twice.
         """
         self.run_coro(self.juliet.send_files([attachment]))
-        self.http_upload.assert_called_with(**upload_kwargs)
+        self.http_upload.assert_called_with(*upload_args)
         self._assert_file()
         self.http_upload.reset_mock()
         self.run_coro(self.juliet.send_files([attachment]))
@@ -196,58 +202,33 @@ class TestAttachmentUpload(Base):
     def test_path(self):
         self.__test_basic(
             LegacyAttachment(path=self.avatar_path),
-            dict(
-                filename=self.avatar_path,
-                content_type="image/png",
-                ifrom=self.xmpp.boundjid,
-                domain=None,
-            ),
+            (self.avatar_path, None, "image/png"),
         )
 
     def test_path_attachment(self):
         self.__test_basic(
             LegacyAttachment(path=self.avatar_path, disposition="attachment"),
-            dict(
-                filename=self.avatar_path,
-                content_type="image/png",
-                ifrom=self.xmpp.boundjid,
-                domain=None,
-            ),
+            (self.avatar_path, None, "image/png"),
             "attachment",
         )
 
     def test_thumbhash(self):
         self.__test_basic(
             LegacyAttachment(path=self.avatar_path, content_type="image/png"),
-            dict(
-                filename=self.avatar_path,
-                content_type="image/png",
-                ifrom=self.xmpp.boundjid,
-                domain=None,
-            ),
+            (self.avatar_path, None, "image/png"),
         )
 
     def test_path_and_id(self):
         self._test_reuse(
             LegacyAttachment(path=self.avatar_path, legacy_file_id=1235),
-            dict(
-                filename=self.avatar_path,
-                content_type="image/png",
-                ifrom=self.xmpp.boundjid,
-                domain=None,
-            ),
+            (ANY, None, "image/png"),
         )
 
     def test_bytes(self):
         with patch("pathlib.Path.stat", return_value=os.stat(self.avatar_path)):
             self.__test_basic(
                 LegacyAttachment(data=self.avatar_path.read_bytes(), name="5x5.png"),
-                dict(
-                    filename=ANY,
-                    content_type="image/png",
-                    ifrom=self.xmpp.boundjid,
-                    domain=None,
-                ),
+                (ANY, "5x5.png", "image/png"),
             )
 
     def test_bytes_and_id(self):
@@ -258,12 +239,7 @@ class TestAttachmentUpload(Base):
                     legacy_file_id=123,
                     name="5x5.png",
                 ),
-                dict(
-                    filename=ANY,
-                    content_type="image/png",
-                    ifrom=self.xmpp.boundjid,
-                    domain=None,
-                ),
+                (ANY, "5x5.png", "image/png"),
             )
 
     def test_race_condition(self):
@@ -282,6 +258,20 @@ class TestAttachmentUpload(Base):
         _msg2 = self.next_sent()
         assert self.next_sent() is None
 
+    def test_fail(self):
+        self.run_coro(self.juliet.send_files([LegacyAttachment(name="chaton.jpg", url="fail", caption="trop mimi")]))
+        self.send(  # language=XML
+            """
+            <message xmlns="jabber:component:accept"
+                     type="chat"
+                     from="juliet@aim.shakespeare.lit/slidge"
+                     to="romeo@montague.lit">
+              <body>/me tried to send a file (chaton.jpg: trop mimi), but something went wrong: IT FAILED.</body>
+            </message>
+            """
+        )
+
+
 class TestAttachmentNoUpload(Base):
     @classmethod
     def setUpClass(cls):
@@ -359,8 +349,9 @@ class TestAttachmentNoUpload(Base):
         assert self.next_sent() is None
         assert len(set(xmpp_ids)) == len(xmpp_ids)
         self.juliet.react(6666, "♥")
-        reaction = self.next_sent()
-        assert reaction["reactions"]["id"] in xmpp_ids
+        for _ in range(len(xmpp_ids)):
+            reaction = self.next_sent()
+            assert reaction["reactions"]["id"] in xmpp_ids
 
         self.recv(  # language=XML
             """
@@ -401,6 +392,120 @@ class TestAttachmentNoUpload(Base):
         self.xmpp.use_message_ids = False
         self.xmpp.LEGACY_MSG_ID_TYPE = True
 
+    def test_multiatt_mark_all_messages_body_caption(self):
+        self.xmpp.MARK_ALL_MESSAGES = True
+        self.run_coro(
+            self.juliet.send_files(
+                [
+                    LegacyAttachment(path=self.avatar_path),
+                ],
+                legacy_msg_id="legacy-real-id",
+            )
+        )
+        assert self.next_sent() is not None
+        self.run_coro(
+            self.juliet.send_files(
+                [
+                    LegacyAttachment(path=self.avatar_path),
+                ],
+                legacy_msg_id="legacy-real-id",
+            )
+        )
+        assert self.next_sent() is not None
+        self.xmpp.MARK_ALL_MESSAGES = False
+
+
+    def test_multi_from_user(self):
+        self.xmpp["xep_0356"].granted_privileges["montague.lit"].message = MessagePermission.OUTGOING
+        self.run_coro(self.juliet.send_files([], body="test", legacy_msg_id="test-msg-id", carbon=True))
+        self.send(  # language=XML
+            """
+            <message xmlns="jabber:component:accept"
+                     to="montague.lit"
+                     from="aim.shakespeare.lit">
+              <privilege xmlns="urn:xmpp:privilege:2">
+                <forwarded xmlns="urn:xmpp:forward:0">
+                  <message xmlns="jabber:client"
+                           type="chat"
+                           from="romeo@montague.lit"
+                           id="test-msg-id"
+                           to="juliet@aim.shakespeare.lit">
+                    <body>test</body>
+                    <active xmlns="http://jabber.org/protocol/chatstates" />
+                    <store xmlns="urn:xmpp:hints" />
+                    <markable xmlns="urn:xmpp:chat-markers:0" />
+                  </message>
+                </forwarded>
+              </privilege>
+            </message>
+            """,
+            # use_values=False,
+        )
+
+        self.juliet.react(legacy_msg_id="test-msg-id", emojis=["🏥"])
+        self.send(  # language=XML
+            """
+            <message type="chat"
+                     from="juliet@aim.shakespeare.lit/slidge"
+                     to="romeo@montague.lit">
+              <store xmlns="urn:xmpp:hints" />
+              <reactions xmlns="urn:xmpp:reactions:0"
+                         id="test-msg-id">
+                <reaction>🏥</reaction>
+              </reactions>
+            </message>
+            """,
+            use_values=False,
+        )
+
+
+    def test_multi_from_user_muc(self):
+        session = self.get_romeo_session()
+        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
+        muc.add_user_resource("gajim")
+        user_part = self.run_coro(muc.get_user_participant())
+        self.run_coro(user_part.send_files([], body="test", legacy_msg_id="test-msg-id"))
+        self.send(  # language=XML
+            """
+            <message type="groupchat"
+                     from="room@aim.shakespeare.lit/romeo"
+                     id="test-msg-id"
+                     to="romeo@montague.lit/gajim">
+              <body>test</body>
+              <active xmlns="http://jabber.org/protocol/chatstates" />
+              <markable xmlns="urn:xmpp:chat-markers:0" />
+              <stanza-id xmlns="urn:xmpp:sid:0"
+                         id="test-msg-id"
+                         by="room@aim.shakespeare.lit" />
+              <occupant-id xmlns="urn:xmpp:occupant-id:0"
+                           id="slidge-user" />
+            </message>
+            """,
+            use_values=False,
+        )
+        with unittest.mock.patch("uuid.uuid4") as uuid:
+            uuid.return_value = "prout"
+            user_part.react(legacy_msg_id="test-msg-id", emojis=["🏥"])
+            self.send(  # language=XML
+                """
+            <message type="groupchat"
+                     from="room@aim.shakespeare.lit/romeo"
+                     to="romeo@montague.lit/gajim">
+              <store xmlns="urn:xmpp:hints" />
+              <stanza-id xmlns="urn:xmpp:sid:0"
+                         id="prout"
+                         by="room@aim.shakespeare.lit" />
+              <reactions xmlns="urn:xmpp:reactions:0"
+                         id="test-msg-id">
+                <reaction>🏥</reaction>
+              </reactions>
+              <occupant-id xmlns="urn:xmpp:occupant-id:0"
+                           id="slidge-user" />
+            </message>
+            """,
+                use_values=False,
+            )
+
     def test_multi_moderation(self):
         session = self.get_romeo_session()
         muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
@@ -429,6 +534,98 @@ class TestAttachmentNoUpload(Base):
             moderated_ids.append(stanza["retract"]["id"])
         assert set(stanza_ids) == set(moderated_ids)
 
+    def test_multi_retraction(self):
+        session = self.get_romeo_session()
+        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
+        muc.add_user_resource("gajim")
+        part = muc.get_system_participant()
+        self.run_coro(
+            part.send_files(
+                [
+                    LegacyAttachment(path=self.avatar_path),
+                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
+                ],
+                legacy_msg_id="the-real-msg-id",
+                body="BODY",
+            )
+        )
+        stanza_ids = []
+        while (stanza := self.next_sent()) is not None:
+            stanza_ids.append(stanza["stanza_id"]["id"])
+        assert len(stanza_ids) == 4  # 2 attachments, the caption and the body
+        assert "the-real-msg-id" in stanza_ids
+
+        part.retract("the-real-msg-id")
+
+        retracted_ids = []
+        while (stanza := self.next_sent()) is not None:
+            retracted_ids.append(stanza["retract"]["id"])
+        assert set(stanza_ids) == set(retracted_ids)
+
+    def test_multi_reply(self):
+        session = self.get_romeo_session()
+        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
+        muc.add_user_resource("gajim")
+        part = muc.get_system_participant()
+        self.run_coro(
+            part.send_files(
+                [
+                    LegacyAttachment(path=self.avatar_path),
+                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
+                ],
+                legacy_msg_id=42,
+                body="BODY",
+            )
+        )
+        stanza_ids = []
+        while (stanza := self.next_sent()) is not None:
+            stanza_ids.append(stanza["stanza_id"]["id"])
+        assert len(stanza_ids) == 4  # 2 attachments, the caption and the body
+
+        for stanza_id in stanza_ids:
+            with unittest.mock.patch("test_shakespeare.Session.on_text") as on_text:
+                self.recv(  # language=XML
+                    f"""
+            <message from='{session.user_jid}/gajim'
+                     to='{muc.jid}'
+                     type='groupchat'>
+              <body>Great idea!</body>
+              <reply to='{part.jid}'
+                     id='{stanza_id}'
+                     xmlns='urn:xmpp:reply:0' />
+            </message>
+            """
+                )
+                on_text.assert_awaited_once()
+                assert on_text.call_args[1].get("reply_to_msg_id") == 42
+
+    def test_multi_react(self):
+        session = self.get_romeo_session()
+        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
+        muc.add_user_resource("gajim")
+        part = muc.get_system_participant()
+        self.run_coro(
+            part.send_files(
+                [
+                    LegacyAttachment(path=self.avatar_path),
+                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
+                ],
+                legacy_msg_id=42,
+                body="BODY",
+            )
+        )
+        stanza_ids = []
+        while (stanza := self.next_sent()) is not None:
+            stanza_ids.append(stanza["stanza_id"]["id"])
+
+        part.react(legacy_msg_id=42, emojis=["♥"])
+
+        reacted_stanzas_ids = []
+        while (stanza := self.next_sent()) is not None:
+            reacted_stanzas_ids.append(stanza["reactions"]["id"])
+
+        assert set(stanza_ids) == set(reacted_stanzas_ids)
+
     def test_reply_with_attachment(self):
         self.run_coro(
             self.juliet.send_files(
