@@ -1,5 +1,6 @@
 import re
-from datetime import datetime, timedelta, timezone
+from datetime import datetime, timezone
+from pathlib import Path
 
 from slixmpp import JID
 
@@ -11,7 +12,7 @@ from slidge.util import (
     SubclassableOnce,
     is_valid_phone_number,
 )
-from slidge.util.types import Mention
+from slidge.util.types import Mention, LegacyAttachment
 from slidge.util.util import merge_resources, replace_mentions, strip_leading_emoji
 
 
@@ -263,3 +264,17 @@ def test_last_seen_fallback_formatting():
     res, should_update = get_last_seen_fallback(now)
     assert should_update
     assert "AM" in res or "PM" in res
+
+
+def test_attachment_str():
+    att = LegacyAttachment(path="whatevs")
+    assert str(att) == f"Attachment(path={Path('whatevs')!r})"
+
+    att = LegacyAttachment(url="https://url", legacy_file_id="xxx")
+    assert str(att) == "Attachment(legacy_file_id='xxx', url='https://url')"
+
+    att = LegacyAttachment(data=b"aaaaa", legacy_file_id="xxx")
+    assert str(att) == "Attachment(legacy_file_id='xxx', data=<5 bytes>)"
+
+    att = LegacyAttachment(data=b"aaaaa")
+    assert str(att) == "Attachment(data=<5 bytes>)"
